# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals
from future import standard_library
standard_library.install_aliases()
from builtins import str

import click
import csv
import os
import urllib.request, urllib.error, urllib.parse

from PIL import Image
from urllib.parse import urljoin

from kraken import binarization
from kraken import pageseg
from kraken import rpred
from kraken import html
from kraken.lib import models

APP_NAME = 'kraken'
MODEL_URL = 'http://l.unchti.me/'
DEFAULT_MODEL = 'en-default.pyrnn.hdf5'
LEGACY_MODEL_DIR = '/usr/local/share/ocropus'


@click.group()
def cli():
    pass


@click.command('binarize')
@click.option('--threshold', default=0.5, type=click.FLOAT)
@click.option('--zoom', default=0.5, type=click.FLOAT)
@click.option('--escale', default=1.0, type=click.FLOAT)
@click.option('--border', default=0.1, type=click.FLOAT)
@click.option('--perc', default=80, type=click.IntRange(1, 100))
@click.option('--range', default=20, type=click.INT)
@click.option('--low', default=5, type=click.IntRange(1, 100))
@click.option('--high', default=90, type=click.IntRange(1, 100))
@click.argument('input', type=click.File(mode='rb'))
@click.argument('output', type=click.File(mode='wb'))
@click.pass_context
def binarize(ctx, threshold, zoom, escale, border, perc, range, low, high,
             input, output):
    im = Image.open(input)
    res = binarization.nlbin(im, threshold, zoom, escale, border, perc, range,
                             low, high)
    res.save(output)


@click.command('segment')
@click.option('--scale', default=None, type=click.FLOAT)
@click.option('-b/-w', '--black_colseps/--white_colseps', default=False)
@click.argument('input', type=click.File(mode='rb'))
@click.argument('output', type=click.File(mode='w', encoding='utf-8'), required=False)
@click.pass_context
def segment(ctx, scale, black_colseps, input, output):
    try:
        im = Image.open(input)
    except IOError as e:
        click.BadParameter(e.message)
    res = pageseg.segment(im, scale, black_colseps)
    for box in res:
        click.echo(u','.join([str(c) for c in box]) + u'\n', file=output,
                   nl=False)


@click.command('ocr')
@click.pass_context
@click.option('-m', '--model', default=DEFAULT_MODEL, help='Path to an '
              'recognition model')
@click.option('-p', '--pad', type=click.INT, default=16, help='Left and right '
              'padding around lines')
@click.option('-h/-t', '--hocr/--text', default=False, help='Switch between '
              'hOCR and plain text output')
@click.option('-l', '--lines', type=click.File(mode='r', encoding='utf-8'),
              required=True, help='JSON file containing line coordinates')
@click.option('--enable-autoconversion/--disable-autoconversion', 'conv',
              default=True, help='Automatically convert pyrnn models zu HDF5')
@click.argument('input', type=click.File(mode='rb'))
@click.argument('output', type=click.File(mode='w', encoding='utf-8'),
                required=False)
def ocr(ctx, model, pad, hocr, lines, conv, input, output):
    # first we try to find the model in the absolue path, then ~/.kraken, then
    # LEGACY_MODEL_DIR
    search = [model,
              os.path.join(click.get_app_dir(APP_NAME, force_posix=True), model),
              os.path.join(LEGACY_MODEL_DIR, model)]
    # if automatic conversion is enabled we look for an converted model in
    # ~/.kraken
    if conv:
        search.insert(0, os.path.join(click.get_app_dir(APP_NAME,
                       force_posix=True),
                       os.path.basename(os.path.splitext(model)[0]) +
                       '.hdf5'))
    location = None
    for loc in search:
        if os.path.isfile(loc):
            location = loc
            break
    if not location:
        raise click.BadParameter('No model found')
    click.echo('Loading RNN\t', nl=False)
    rnn = models.load_any(location)
    if rnn:
        click.secho(u'\u2713', fg='green')
    else:
        click.secho(u'\u2717', fg='red')
        ctx.exit(1)
    # convert input model to HDF5
    if conv and rnn.kind == 'pyrnn':
        name, _ = os.path.splitext(os.path.basename(model))
        op = os.path.join(click.get_app_dir(APP_NAME, force_posix=True), name + '.hdf5')
        try:
            os.makedirs(click.get_app_dir(APP_NAME, force_posix=True))
        except OSError:
            pass
        models.pyrnn_to_hdf5(rnn, op)
    try:
        im = Image.open(input)
    except IOError as e:
        click.BadParameter(e.message)
    lc = len(lines.readlines())
    lines.seek(0)
    with click.progressbar(csv.reader(lines), lc, label='Reading line bounds',
                           fill_char=click.style('#', fg='green'),) as b:
        bounds = [(int(x1), int(y1), int(x2), int(y2)) for x1, y1, x2, y2 in b]

    it = rpred.rpred(rnn, im, bounds, pad)
    with click.progressbar(it, len(bounds),
                           label='Recognizing lines',
                           fill_char=click.style('#', fg='green')) as pred:
        records = []
        for rec in pred:
            records.append(rec)
        if hocr:
            click.echo(html.hocr(records, input.name, im.size), file=output,
                       nl=False)
        else:
            click.echo(u'\n')
            click.echo(u'\n'.join([s.prediction for s in records]),
                       file=output, nl=False)


@click.command('download')
def download():
    default_model = urllib.request.urlopen(urljoin(MODEL_URL, DEFAULT_MODEL))
    try:
        os.makedirs(click.get_app_dir(APP_NAME, force_posix=True))
    except OSError:
        pass
    # overwrite next function for iterator to return 8192 octets instead of
    # line
    default_model.next = lambda: default_model.read(8192)
    fs = int(default_model.info()["Content-Length"])
    with open(os.path.join(click.get_app_dir(APP_NAME, force_posix=True),
                           DEFAULT_MODEL), 'wb') as fp:
        with click.progressbar(default_model, fs // 256,
                               label='Downloading default model',
                               fill_char=click.style('#', fg='green')) as dl:
            for buf in dl:
                if not buf:
                    raise StopIteration()
                fp.write(buf)


cli.add_command(binarize)
cli.add_command(segment)
cli.add_command(ocr)
cli.add_command(download)



from __future__ import absolute_import, division, print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()
from builtins import range
from builtins import object

import pickle
import gzip
import bz2
import sys
import numpy as np

from kraken.lib import lstm
from kraken.lib.util import pil2array, array2pil
from kraken.lib.lineest import CenterNormalizer
from kraken.lib.exceptions import KrakenInvalidModelException, KrakenInputException


class ocr_record(object):
    """
    A record object containing the recognition result of a single line
    """
    def __init__(self, prediction, cuts, confidences):
        self.prediction = prediction
        self.cuts = cuts
        self.confidences = confidences

    def __len__(self):
        return len(self.prediction)

    def __str__(self):
        return self.prediction

    def __iter__(self):
        self.idx = -1
        return self

    def __next__(self):
        if self.idx + 1 < len(self):
            self.idx += 1
            return (self.prediction[self.idx], self.cuts[self.idx],
                    self.confidences[self.idx])
        else:
            raise StopIteration

    def __getitem__(self, key):
        if isinstance(key, slice):
            return [self[i] for i in range(*key.indices(len(self)))]
        elif isinstance(key, int):
            if key < 0:
                key += len(self)
            if key >= len(self):
                raise IndexError('Index (%d) is out of range' % key)
            return (self.prediction[key], self.cuts[key],
                    self.confidences[key])
        else:
            raise TypeError('Invalid argument type')


def extract_boxes(im, bounds):
    """
    Yields the subimages of image im defined in the list of bounding boxes in
    bounds preserving order.

    Args:
        im (PIL.Image): Input image
        bounds (list): A list of tuples (x1, y1, x2, y2)

    Yields:
        (PIL.Image) the extracted subimage
    """
    for box in bounds:
        if (box < (0, 0, 0, 0) or box[::2] > (im.size[0], im.size[0]) or
            box[1::2] > (im.size[1], im.size[1])):
            raise KrakenInputException('Line outside of image bounds')
        yield im.crop(box), box


def dewarp(normalizer, im):
    """
    Dewarps an image of a line using a kraken.lib.lineest.CenterNormalizer
    instance.

    Args:
        normalizer (kraken.lib.lineest.CenterNormalizer): A line normalizer
                                                          instance
        im (PIL.Image): Image to dewarp

    Returns:
        PIL.Image containing the dewarped image.
    """
    line = pil2array(im)
    temp = np.amax(line)-line
    temp = temp*1.0/np.amax(temp)
    normalizer.measure(temp)
    line = normalizer.normalize(line, cval=np.amax(line))
    return array2pil(line)


def rpred(network, im, bounds, pad=16, line_normalization=True):
    """
    Uses a RNN to recognize text

    Args:
        network (kraken.lib.lstm.SegRecognizer): A SegRecognizer object
        im (PIL.Image): Image to extract text from
        bounds (iterable): An iterable returning a tuple defining the absolute
                           coordinates (x0, y0, x1, y1) of a text line in the
                           Image.
        pad (int): Extra blank padding to the left and right of text line
        line_normalization (bool): Dewarp line using the line estimator
                                   contained in the network. If no normalizer
                                   is available one using the default
                                   parameters is created. By aware that you may
                                   have to scale lines manually to the target
                                   line height if disabled.
    Yields:
        A tuple containing the recognized text (0), absolute character
        positions in the image (1), and confidence values for each
        character(2).
    """

    lnorm = getattr(network, 'lnorm', CenterNormalizer())

    for box, coords in extract_boxes(im, bounds):
        if line_normalization:
            box = dewarp(lnorm, box)
        line = pil2array(box)
        raw_line = line.copy()
        line = lstm.prepare_line(line, pad)
        pred = network.predictString(line)

        # calculate recognized LSTM locations of characters
        scale = len(raw_line.T)/(len(network.outputs)-2 * pad)
        result = lstm.translate_back(network.outputs, pos=1)
        pos = [(coords[0], coords[1], coords[0], coords[3])]
        conf = []
        for r, c in result:
            pos.append((pos[-1][2], coords[1],
                        coords[0] + int((r-pad) * scale),
                        coords[3]))
            conf.append(network.outputs[r, c])
        yield ocr_record(pred, pos[1:], conf)



"""
kraken.lib.models
~~~~~~~~~~~~~~~~~

Wraps around legacy pyrnn and HDF5 models to provide a single interface. In the
future it will also include support for clstm models.
"""

from __future__ import absolute_import
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()
from future.utils import PY3
from builtins import next
from builtins import chr

import h5py
import numpy
import gzip
import bz2
import sys

import kraken.lib.lstm
import kraken.lib.lineest

from kraken.lib.lineest import CenterNormalizer
from kraken.lib.exceptions import KrakenInvalidModelException

def load_any(fname):
    """
    Loads anything that was, is, and will be a valid ocropus model and
    instantiates a shiny new kraken.lib.lstm.SeqRecognizer from the RNN
    configuration in the file.
    
    Currently it recognizes the following kinds of models:
        * pyrnn models containing BIDILSTMs
        * HDF5 models containing converted python BIDILSTMs

    Additionally an attribute 'kind' will be added to the SeqRecognizer
    containing a string representation of the source kind. Current known values
    are:
        * pyrnn for pickled BIDILSTMs
        * hdf-pyrnn for HDF models converted from pickled objects

    Args:
        fname (unicode): Path to the model

    Returns:
        A kraken.lib.lstm.SeqRecognizer object.
    """
    seq = None
    try:
        seq = load_hdf5(fname)
        seq.kind = 'hdf-pyrnn'
        return seq
    except:
        try:
            seq = load_pyrnn(fname)
            seq.kind = 'pyrnn'
            return seq
        except:
            raise 


def load_hdf5(fname, line_height=0):
    """
    Loads a model in HDF5 format and instantiates a
    kraken.lib.lstm.SeqRecognizer object.

    Args:
        fname (unicode): Path to the HDF5 file
        line_height (int): Target height of input. Will be extracted from model
        if set to 0.

    Returns:
        A kraken.lib.lstm.SeqRecognizer object
    """
    known_hdf5 = [b'pyrnn-bidi']
    with h5py.File(fname, 'r') as rnn:
        try: 
            if rnn.attrs['kind'] not in known_hdf5:
                raise KrakenInvalidModelException(b'Unknown model kind ' +
                                                  rnn.attrs['kind'])
            # first extract the codec character set
            charset = [chr(x) for x in rnn.get('codec')]
            # code 0 is handled separately by the model
            charset[0] = ''
            codec = kraken.lib.lstm.Codec().init(charset)

            # get number of states from shape of first array
            hiddensize = rnn['.bidilstm.0.parallel.0.lstm.WGI'].shape[0]
            if not line_height:
                line_height = rnn['.bidilstm.0.parallel.0.lstm.WGI'].shape[1] - hiddensize - 1

            # next build a line estimator
            lnorm = kraken.lib.lineest.CenterNormalizer(line_height)
            network = kraken.lib.lstm.SeqRecognizer(lnorm.target_height, 
                                         hiddensize, 
                                         codec = codec, 
                                         normalize = kraken.lib.lstm.normalize_nfkc)
            parallel, softmax = network.lstm.nets
            nornet, rev = parallel.nets
            revnet = rev.net
            for w in ('WGI', 'WGF', 'WGO', 'WCI', 'WIP', 'WFP', 'WOP'):
                setattr(nornet, w, rnn['.bidilstm.0.parallel.0.lstm.' +
                        w][:].reshape(getattr(nornet, w).shape))
                setattr(revnet, w, rnn['.bidilstm.0.parallel.1.reversed.0.lstm.' +
                        w][:].reshape(getattr(nornet, w).shape))
            softmax.W2 = numpy.hstack((rnn['.bidilstm.1.softmax.w'][:],
                                       rnn['.bidilstm.1.softmax.W'][:]))
        except (KeyError, TypeError):
            raise KrakenInvalidModelException('Model incomplete')
        network.lnorm = lnorm
        return network

def load_pyrnn(fname):
    """
    Loads a legacy RNN from a pickle file.

    Args:
        fname (unicode): Path to the pickle object

    Returns:
        Unpickled object

    """
    
    if PY3:
        raise KrakenInvalidModelException('Loading pickle models is not '
                                          'supported on python 3')
    import cPickle
    def find_global(mname, cname):
        aliases = {
            'lstm.lstm': kraken.lib.lstm,
            'ocrolib.lstm': kraken.lib.lstm,
            'ocrolib.lineest': kraken.lib.lineest,
        }
        if mname in aliases:
            return getattr(aliases[mname], cname)
        return getattr(sys.modules[mname], cname)

    of = open
    if fname.endswith(u'.gz'):
        of = gzip.open
    elif fname.endswith(u'.bz2'):
        of = bz2.BZ2File
    with of(fname, 'rb') as fp:
        unpickler = cPickle.Unpickler(fp)
        unpickler.find_global = find_global
        try:
            rnn = unpickler.load()
        except cPickle.UnpicklingError as e:
            raise KrakenInvalidModelException(str(e))
        if not isinstance(rnn, kraken.lib.lstm.SeqRecognizer):
            raise KrakenInvalidModelException('Pickle is %s instead of '
                                              'SeqRecognizer' %
                                              type(rnn).__name__)
        return rnn

def pyrnn_to_hdf5(pyrnn=None, output='en-default.hdf5'):
    """
    Converts a legacy python RNN to the new HDF5 format. Benefits of the new
    format include independence from particular python versions and no
    arbitrary code execution issues inherent in pickle.

    Args:
        pyrnn (kraken.lib.lstm.SegRecognizer): pyrnn model
        output (unicode): path of the converted HDF5 model
    """

    parallel, softmax = pyrnn.lstm.nets
    fwdnet, revnet = parallel.nets
    
    with h5py.File(output, 'w') as nf:
        # write metadata first
        nf.attrs['kind'] = 'pyrnn-bidi'

        for w in ('WGI', 'WGF', 'WGO', 'WCI'):
            dset = nf.create_dataset(".bidilstm.0.parallel.0.lstm." + w,
                                     getattr(fwdnet, w).shape, dtype='f')
            dset[...] = getattr(fwdnet, w)
            dset = nf.create_dataset(".bidilstm.0.parallel.1.reversed.0.lstm." + w,
                                     getattr(revnet.net, w).shape, dtype='f')
            dset[...] = getattr(revnet.net, w)

        for w in ('WIP', 'WFP', 'WOP'):
            data = getattr(fwdnet, w).reshape((-1, 1))
            dset = nf.create_dataset(".bidilstm.0.parallel.0.lstm." + w,
                                     data.shape, dtype='f')
            dset[...] = data
            
            data = getattr(revnet.net, w).reshape((-1, 1))
            dset = nf.create_dataset(".bidilstm.0.parallel.1.reversed.0.lstm." + w,
                                     data.shape, dtype='f')
            dset[...] = data

        dset = nf.create_dataset(".bidilstm.1.softmax.w",
                                 (softmax.W2[:,0].shape[0], 1), dtype='f')
        dset[:] = softmax.W2[:,0].reshape((-1, 1))

        dset = nf.create_dataset(".bidilstm.1.softmax.W", softmax.W2[:,1:].shape,
                                 dtype='f')
        dset[:] = softmax.W2[:,1:]
        cvals = iter(pyrnn.codec.code2char.values())
        next(cvals)
        codec = numpy.array([0]+[ord(x) for x in cvals], dtype='f').reshape((-1, 1))
        dset = nf.create_dataset("codec", codec.shape, dtype='f')
        dset[:] = codec



# -*- coding: utf-8 -*-
import unittest
import os
import tempfile
import pickle
import h5py

from nose.tools import raises 

from kraken.lib import models
from kraken.lib.exceptions import KrakenInvalidModelException

class TestModels(unittest.TestCase):

    """
    Testing model loading routines
    """

    def setUp(self):
        self.temp = tempfile.NamedTemporaryFile(delete=False)

    def tearDown(self):
        self.temp.close()
        os.unlink(self.temp.name)

    @raises(KrakenInvalidModelException)
    def test_load_hdf5_unknown(self):
        """
        Test correct handling of unknown HDF5 files.
        """
        b = h5py.File(self.temp.name, 'w')
        b.attrs['kind'] = 'not-an-rnn'
        b.close()
        models.load_hdf5(self.temp.name)

    @raises(KrakenInvalidModelException)
    def test_load_hdf5_incomplete(self):
        """
        Test correct handling of incomplete HDF5 files.
        """
        b = h5py.File(self.temp.name, 'w')
        b.attrs['kind'] = 'pyrnn-bidi'
        b.close()
        models.load_hdf5(self.temp.name)

    @raises(IOError)
    def test_load_hdf5_invalid(self):
        """
        Test correct handling of invalid HDF5 files.
        """
        models.load_hdf5(self.temp.name)
       
    @raises(KrakenInvalidModelException)
    def test_load_pyrnn_invalid(self):
        """
        Test correct handling of non-pickle files.
        """
        b = h5py.File(self.temp.name, 'w')
        b.attrs['kind'] = 'pyrnn-bidi'
        b.close()
        models.load_pyrnn(self.temp.name)

    @raises(KrakenInvalidModelException)
    def test_load_pyrnn_no_seqrecognizer(self):
        """
        Test correct handling of non-SeqRecognizer pickles.
        """
        pickle.dump(u'Iámnõtãrécðçnízer', self.temp)
        self.temp.close()
        models.load_pyrnn(self.temp.name)

    def test_load_pyrnn_gz(self):
        """
        Test correct handling of gzipped models.
        """

    def test_load_pyrnn_bz2(self):
        """
        Test correct handling of bzip2 compressed models.
        """
        pass

    def test_load_pyrnn_uncompressed(self):
        """
        Test correct handling of uncompressed models.
        """
        pass

    def test_load_pyrnn_aliasing(self):
        """
        Test correct aliasing of ocrolib classes.
        """
        pass

    def test_load_pyrnn_aliasing_old(self):
        """
        Test correct aliasing of pre-ocrolib classes.
        """
        pass
